
#include "inc.h"

#include "minix.h"
#include "irq_arm.h"
#include "bios_arm.h"
#include "div_arm.h"
#include "gbfs.h"
#include "lib.h"

#include "screen.h"
#include "sudoku.h"

//---------------------------------------------------------------------------
ST_SUDOKU Sudoku;

//---------------------------------------------------------------------------
EWRAM_CODE void SudokuInit()
{
	_Memset((u8*)&Sudoku, 0x00, sizeof(ST_SUDOKU));

	u16 sprNo = 1;
	u16 x, y;

	for(y=0; y<CELL_LEN_Y; y++)
	{
		for(x=0; x<CELL_LEN_X; x++)
		{
			SpriteSetSize(sprNo, SP_SIZE_16, SP_SQUARE, SP_COLOR_16);
			SpriteMove(sprNo, CELL_SX+CELL_CX*x, CELL_SY+CELL_CY*y);

			sprNo++;
		}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE bool SudokuLoadStage(u8* pDat)
{
	u8  chr;
	u16 x, y;
	u16 i;

	for(i=0; i<CELL_MAX_CNT; i++)
	{
		x   = _UMod(i, CELL_LEN_X);
		y   = _UDiv(i, CELL_LEN_X);
		chr = *pDat++;

		if( _IsDigit(chr) == TRUE )
		{
			SudokuSetCell(x, y, chr - 0x30);
		}
		else
		{
			return FALSE;
		}
	}

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void SudokuSetCell(u16 x, u16 y, u16 num)
{
	ST_SOLVE_CELL* b = &Sudoku.solveCell[x][y];
	ST_CELL*       p = &Sudoku.cell[x][y];

	p->type = (num == 0) ? TYPE_VAR : TYPE_CONST;
	p->num  = num;

	b->num  = num;
}
//---------------------------------------------------------------------------
EWRAM_CODE void SudokuDrawCellAll()
{
	u16 x, y;

	for(y=0; y<CELL_LEN_Y; y++)
	{
		for(x=0; x<CELL_LEN_X; x++)
		{
			SudokuDrawCell(x, y);
		}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void SudokuDrawCell(u16 x, u16 y)
{
	ST_CELL* p = &Sudoku.cell[x][y];

	u16 sprNo   = CELL_LEN_Y * y + x + 1;
	u16 sprTile = (p->num * 2) + ((p->type == TYPE_CONST) ? 64 : 0);

	SpriteSetChr(sprNo, sprTile + 512);
}
//---------------------------------------------------------------------------
EWRAM_CODE void SudokuDrawBg()
{
	Mode3DrawImage(0, 0, SCN_IMG_CX, SCN_IMG_CY, (u16*)scnBitmap);
}
//---------------------------------------------------------------------------
EWRAM_CODE void SudokuProcSolve()
{
	u8  type;
	u16 x, y;
	s16 cnt = 0;
	s16 dir = 1;

	while(cnt < CELL_LEN_X*CELL_LEN_Y)
	{
		x    = _UMod(cnt, CELL_LEN_X);
		y    = _UDiv(cnt, CELL_LEN_Y);
		type = Sudoku.cell[x][y].type;

		if(type == TYPE_CONST)
		{
			cnt += dir;
			continue;
		}

		if( SudokuSetSolveCell(x, y) == TRUE )
		{
			cnt++;
			dir = 1;
		}
		else
		{
			cnt--;
			dir = -1;
		}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE bool SudokuSetSolveCell(u16 x, u16 y)
{
	ST_SOLVE_CELL* p = &Sudoku.solveCell[x][y];

	//LqȂꏊ́Ǎ⌟܂
	if(p->num == 0)
	{
		SudokuSetSolveFlagClear(p);
		SudokuSetSolveFlagVert(p, x);
		SudokuSetSolveFlagHoriz(p, y);
		SudokuSetSolveFlagArea(p, x, y);
	}

	while(p->cnt < 9)
	{
		p->cnt++;

		if(p->flag[ p->cnt ] == FALSE)
		{
			p->num = p->cnt;
			return TRUE;
		}
	}

	p->num = 0;
	p->cnt = 0;
	return FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void SudokuSetSolveFlagClear(ST_SOLVE_CELL* p)
{
	u16 i;

	for(i=0; i<9+1; i++)
	{
		p->flag[i] = FALSE;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void SudokuSetSolveFlagVert(ST_SOLVE_CELL* p, u16 x)
{
	ST_SUDOKU* s = &Sudoku;
	u16 y;

	for(y=0; y<CELL_LEN_Y; y++)
	{
		p->flag[ s->solveCell[x][y].num ] = TRUE;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void SudokuSetSolveFlagHoriz(ST_SOLVE_CELL* p, u16 y)
{
	ST_SUDOKU* s = &Sudoku;
	u16 x;

	for(x=0; x<CELL_LEN_X; x++)
	{
		p->flag[ s->solveCell[x][y].num ] = TRUE;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void SudokuSetSolveFlagArea(ST_SOLVE_CELL* p, u16 x, u16 y)
{
	ST_SUDOKU* s = &Sudoku;

	u16 sx = _UDiv(x, 3) * 3;
	u16 sy = _UDiv(y, 3) * 3;
	u16 i, j;

	for(i=0; i<3; i++)
	{
		for(j=0; j<3; j++)
		{
			p->flag[ s->solveCell[sx+i][sy+j].num ] = TRUE;
		}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void SudokuChgCell(u16 x, u16 y, s16 add)
{
	ST_CELL* p = &Sudoku.cell[x][y];

	if(p->type == TYPE_CONST)
	{
		return;
	}

	p->num += add;
	if(p->num >= 10)
	{
		p->num = 1;
	}
	else if(p->num < 0)
	{
		p->num = 9;
	}

	SudokuDrawCell(x, y);
}
//---------------------------------------------------------------------------
EWRAM_CODE bool SudokuIsGameEnd()
{
	u16 x, y;

	for(y=0; y<CELL_LEN_Y; y++)
	{
		for(x=0; x<CELL_LEN_X; x++)
		{
			if(Sudoku.cell[x][y].num != Sudoku.solveCell[x][y].num)
			{
				return FALSE;
			}
		}
	}

	return TRUE;
}
